#!/local/paths/perl -wT
#
# File: newsyslog.pl
#
# Description: Configurable log file rotator written in Perl.
#
#              Reads $config_file to gather information about log files
#              to be rotated. Defaults to keeping an unlimited number of
#              copies of the log file in the same directory.
#
#              This version attempts to improve on previous versions of
#              newsyslog by offering improved error checking, and
#              sending a HUP only to the process associated with the
#              log file that has just moved, rather than HUPping every
#              process that writes to any log file.
#
#              Files which cannot be rotated will cause an error to be
#              raised, and will be skipped.
#
#              Errors are reported to stderr unless the '-q' (quiet)
#              command-line option was selected. If the '-m' (mail)
#              option was selected, errors will be mailed to the
#              specified recipient. Errors are always logged via the
#              syslog facility.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02 (begin)
#
#       1998/02/04: - Change format of config file to split fields on a
#                     colon surrounded by (optional) whitespace.
#                   - More error checking while reading config file.
#                   - Comment lines can now contain whitespace before
#                     the leading #.
#                   - Optional field in config file for backup file
#                     name.
#                   - Add logging via syslog. Facility to use is defined
#                     in configuration section.
#                   - Add info_message routine to print informational
#                     messages when not running quietly.
#                   - All message routines, (fatal_error, error_message,
#                     debug_message, and info_message) log via syslog
#                     with (hopefully) appropriate priority.
#                   - Add '-m recipient' option to send errors to
#                     recipient. Do not mail errors to anyone if not
#                     set.
#
#       1998/02/06: - Compression is now configurable on a per file
#                     basis in the config file, so we don't need the
#                     command line option any more.
#                   - %backupfile array (hash) added.
#                   - %compress   array (hash) added.
#                   - We default to keeping an unlimited number of
#                     cycles if there is no $cycles field in the config
#                     file.
#                   - We determine whether or not we compress the backup
#                     file.
#                   - Fix bug in errors file name: Use _double_ quotes!
#                   - Other bug fixes in Read_Config.
#                   - debug messages go to syslog only if running in
#                     debug mode.
#                   - Increase verbosity of error messages: Line number
#                     in config file and copy of line. 
#
#       1998/02/09: - Some reorganization, and addition of more checks
#                     in Read_Config subroutine.
#                   - Begin working on Move_Logs subroutine which moves
#                     an existing log file out of the way and creates a
#                     new one, then HUPs the associated process if
#                     appropriate.
#
#       1998/02/10: - More work on Move_Logs.
#                   - Bug fix to ensure blank $cycles is accepted.
#
#       1998/02/12: - Begin working on Rotate_Logs subroutine.
#                   - Configurable list of possible filename extensions
#                     for old backups.
#
#       1998/02/16: - Finish Rotate_Logs.
#                   - write Do_Compress.
#                   - write Clean_up.
#                   - test; some small bug fixes.
#                   - test some more.
#
#       1998/02/17: - Fixes from Anne's comments.
#                   - Spelling, grammar, and organization of comments.
#                   - We don't need to send the program name to the
#                     info_message subroutine, since that is already
#                     being logged by the call to syslog.
#                   - untaint_file accepts '+' in a filename.
#                   - Parse_Options gives a more informative error
#                     message if the path to $config_file (optionally
#                     provided on the command line) is tainted.
#                   - Read_Pid_File is now a routine called $Get_Pid,
#                     which does a whole lot better error checking and
#                     reporting.
#                   - Read_Config no longer manually counts input lines.
#                     The count is already available as $.
#
#       1998/02/18: - More fixes to Read_Config:
#                   - Bug fix to give a more informative error message
#                     in the case that file paths are tainted;
#                   - Error messages indicate if they skip a line;
#                   - File and directory paths must be absolute;
#                   - We complain if we read through the whole config
#                     file and don't find a line matching the selected
#                     frequency.
#                   - The cycles field in the config file now must be a
#                     number or the case insensitive string 'date'.
#                     There is no longer a "blank" setting to provide
#                     default behaviour.
#                   - The 'compress' string is now case insensitive.
#                   - The frequency entry in the config file can now be
#                     surrounded by whitespace.
#                   - The bulk of error checking on the $pid from
#                     $pid_file has been moved to Get_Pid.
#                   - We only read a $pid_file (and test signal the
#                     process) once if successful. Further entries which
#                     refer to the same $pid_file (such as syslog files)
#                     will simply use the $pid we already found.
#                   - The result of not being able to open the config
#                     file has been moved to the top of the if statement
#                     for easier reading.
#                   - Changes in Move_Logs:
#                   - The sym-link unravelling is completely redone.
#                   - no more chdir, since we just unravelled sym-links
#                     to find the real directory anyway.
#                   - We no longer maintain separate lists for
#                     "done_files", "skip_rotate" (could not HUP), etc.
#                     The behaviour now is that the %logfiles lists are
#                     altered if files cannot be moved aside or
#                     processes cannot be HUPped. Essentially the lists
#                     in %logfiles will always be the lists of files
#                     which we are continuing to work on.
#                   - Some clarifications in comments.
#                   - Changed in Rotate_Logs:
#                   - We now only refer to the %logfiles hash for lists
#                     of files to work on.
#                   - Bug fix: No extra adjustment is necessary for year
#                     2000.
#
#       1998/02/19: - More work on Rotate_Logs:
#                   - Much of the code is reorganized.
#                   - Smarter progress/error checking is being done.
#                   - Do_Compress is cleaned up. There is no need to try
#                     and sort out what went wrong if a compression
#                     failed. What needs to be done is to remove a stray
#                     .gz file if one exists in this case.
#                   - There is no longer a need for a Clean_Up
#                     subroutine. Its task is now part of Rotate_Logs.
#                   - There is no longer a need for @file_extensions
#                     list. We accept any possible extension if one
#                     exists, (when rotating previous logs).
#                   - We actually *do* try and mail errors if
#                     appropriate. <*blush*>
#                   - Some more fixes to Do_Compress.
#
#       1998/02/20: - Bug fix: The copy that happens if a hardlink fails
#                     while rotating the log file had the wrong arguments.
#                   - Change this program's view of date. Originally it
#                     was assumed that the program was going to run shortly
#                     after midnight, and therefore require yesterday's
#                     date. This is not the case. The program runs shortly
#                     BEFORE midnight.
#                   - Bug fix: logic error when testing if a new file has
#                     been correctly chowned and chmodded.
#
#       1998/02/22: - Bug fix: Error in call to kill HUP.
#                   - Bug fix: incorrect grouping when checking chmod/chown.
#
#       1998/02/25: - Ver 2.1: Additional 'date' keywords permitted in
#                     the configuration file: 'date-yesterday',
#                     'date-yestermonth', and 'date-time'.
#
#       1998/02/26: - /local/paths added to the PATH environment.
#
#       1998/03/02: - More bug fixes: relative symlinks to directories
#                     were incorrectly handled; and one error message
#                     was incorrect.
#                   - Fixed the newsyslog.pl test syslog message.
#                   - The PATH is now set in the configuration section
#                     for easier location when editing.
#                   - "tainted" error messages now wrap quotes around
#                     the tainted argument, for clarity.
#
#       1998/09/15: - Bug fix: include the $! error string in error
#                     messages when renaming a file fails.
#
#       1998/11/24: - Clean up the appearance of some of the error messages.
#                   - We capture the output (if any) of system commands,
#                     and print it along with our own error message.
#       1999/02/01: - Bug fix in Move_Logs: relative symlinks mishandled,
#                     and symlinked log files not correctly rotated.
#       2000/02/29: - Bug fix: trim surrounding whitespace in
#                     configuration file lines.
#                     Bug fix: be paranoid about creating a unique file
#                     name for the errors file.
#                     Bug fix: $ENV{IFS} = " \t\n";
#                     Bug fix: On some systems, (perl-5.005?), local $.
#                     doesn't appear to protect the global $.
#       2001/04/18: - Improved "rotations complete" info message and a
#                     few other messages by adding $frequency.
#                     Changed the mail program from mailx to sendmail.
#                     Bug fix: would have lost the errors if -m not
#                     specified, or mail program failed on "close".
#                     Changed default location of conf file for
#                     pkg-style installation.
#       2001/04/20: - Added taint checks after "readlink" calls.
#                     Simplified the "break out when too many
#                     symlinks" code using a label on the foreach
#                     statement.
#                     Bug fix: detect an existing gz file *before*
#                     trying to compress a log file.
#       2002/12/03: - Added support for Tru64-5.x's /{memb}/ symlink
#                     targets. (ie.: "/var/adm/wtmp" is a symlink to
#                     "../cluster/members/{memb}/adm/wtmp")
#                   - New option: "-t" for test mode: do not rotate any
#                     files, but read through configuration, reporting
#                     errors, and report which files would be rotated.
#

#
# First some directives for the compiler.
use strict;            # Everybody should use this, IM(NS)HO
use Getopt::Std;       # command line arguments
use File::Basename;    # filename parsing
use Sys::Syslog;       # syslog logging
use Sys::Hostname;     # for subject line


# Environment variables. Keep these to known safe values.
$ENV{PATH} = '/local/paths:/bin:/usr/bin:/usr/sbin';
$ENV{IFS}  = " \t\n";       # <space><tab><newline>


####################################################################
#
# CONFIGURATION section. Edit to your heart's content!

# Absolute path to the configuration file.
my $config_file = '/local/pkg/newsyslog/root-newsyslog-2.4/etc/newsyslog.conf';

# The syslog facility to use if running as uid = 0:
my $root_syslog_facility = 'syslog';

# The syslog facility to use if running as any other user. This may or
# may not be the same as $root_syslog_facility.
my $user_syslog_facility = 'user';

# The program to use for mailing errors. Note that this program must
# find its recipients in the message header.  The recipient is
# determined by our -m command-line option. Without -m, errors
# will not be mailed.
my $mail_program = 'sendmail -t -oi';

# The subject we'll use when mailing errors.
my $subject = hostname();
$subject   =~ s/\..*//;
$subject   .= ": newsyslog ERRORS";

#
# End of configuration.
####################################################################

#
# Global variables which are used throughout

# frequency determines which entries in config file to process.
my $frequency;

# backupdir is associated by $log_file = $backup_dir.
my %backupdir;

# backupfile is associated by $log_file = $backup_file.
my %backupfile;

# rotation is associated by $log_file = $cycles.
my %rotation;

# compress is associated by $log_file = $compression.
my %compress;

# logfiles is associated by $pid = @log_files.
my %logfiles;

# realfiles is associated by $log_file = $realfile.
my %realfiles;

# The time taken when the program starts
my $time_at_start;

# Absolute path to the errors file. So far, an errors file is created
# whether or not it will be mailed to anyone. The file is subsequently
# deleted at the end of the cycle.
#
# 2000/02/29: This file is now created automatically, in order to reduce
#             the risk of a sym-link attack based on predictable filenames.
my $errors_file;


#
# "Option" variables.
use vars qw($opt_d $opt_f $opt_m $opt_q $opt_t);

# They translate to:
my $debug     = 0;
my $mail_rcpt = 0;
my $quiet     = 0;
my $test_mode = 0;

# If set, $opt_f overrides $config_file default.


# SUBROUTINE: fatal_error( $message )
#
# Description: executes a call to die, with $message as an
#              argument.
#
# NOTE: We want to syslog the error message before dying.
#
# Author: Sylvain Robitaille
#
# Date: June 19, 1997
#
#       1998/02/04: include logging via syslog.
#
sub fatal_error {
    syslog 'err', "FATAL error: @_\n";
    die "FATAL error: @_\n";
}

# SUBROUTINE: debug_message( $message )
#
# Description: Prints $message to STDOUT, clearly
#              identified as debug information. Prints only 
#              if $debug is a true value.
#
# Author: Sylvain Robitaille
#
# Date: July 22, 1997
#
#       1998/02/04: - include logging via syslog.
#
#       1998/03/06: - but only if in debug mode!
#
sub debug_message {
  if( $debug ) {
    syslog 'debug', "@_\n";
    print "Debug: @_\n";

  open  ERROR,  ">>$errors_file" or fatal_error "can't open $errors_file!";
  print ERROR   "Debug: @_\n";
  close ERROR;
  }
}

# SUBROUTINE: error_message( $message )
#
# Description: Prints messages to $errors_file if writable, and to
#              STDERR if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#
#       1998/02/04: - include logging via syslog.
#
#       1998/02/06: - prefix message with the string 'Error: '
#
sub error_message {
  syslog 'err', "Error: @_\n";
  print STDERR  "Error: @_\n" unless $quiet;

  open  ERROR,  ">>$errors_file" or fatal_error "can't open $errors_file!";
  print ERROR   "Error: @_\n";
  close ERROR;
}

# SUBROUTINE: info_message( $message )
#
# Description: Prints messages to STDOUT if not in quiet mode.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/04
#
sub info_message {
  syslog 'info', "@_\n";
  print STDOUT "@_\n" unless $quiet;
}

# SUBROUTINE: untaint_file( $path )
#
# Description: Ensures that no illegal file characters are
#              in $path.
#
# Returns: An untainted copy of $path or the string 'TAINTED'.
#
# NOTES: This currently checks only that characters
#        contained in $path are legal. A more complete
#        untaint mechanism might check that the file
#        pointed to by $path is a real file, and that the
#        user has access to that file, or that the user
#        can create the file if it doesn't exist, etc.
#
# Author: Sylvain Robitaille
#
# Date: June 19, 1997
#
#       1998/02/04: - rather than call it a fatal error if the path is
#                     tainted, return 0. We will check for this and
#                     simply complain and skip a file if it contains a
#                     tainted path.
#
#       1998/02/06: - return the string 'TAINTED' if the path is
#                     tainted. This will make it easier to check the
#                     return value, by a string comparison rather than
#                     checking for a numeric value.
#
#       1998/02/17: - '+' is a legal filename character.
#
#       2002/12/03: - DEC/OSF 5.1A uses "/{memb}/" in pathnames to denote
#                     cluster-member-specific directories (even on stand-
#                     alone systems, which are effectively one-member
#                     clusters).  It seems as though "/member/" can be
#                     in the same way, so let's translate before checking
#                     for tainted characters.
#
sub untaint_file {
    my $path = shift @_;

    # The following applies only if our OS is "dec_osf"
    if($^O eq 'dec_osf') {
        $path =~ s%/{memb}/%/member/%g;
    }

    if($path =~ /^([\/\-@\w\.~\+]+)$/) {
        $path = $1;
    } else {
        $path = 'TAINTED';
    }
    return $path;
}

# SUBROUTINE: untaint_mail( $path )
#
# Description: Ensures that no illegal file characters are
#              in $path.
#
# Returns: An untainted copy of $path or the string 'TAINTED'.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/19: - copy from untaint_file and adjust.
#
#       1998/02/20: - allow whitespace to permit multiple recipients.
#
sub untaint_mail {
    my $path = shift @_;

    if($path =~ /^([\s\-@\w\.~\+]+)$/) {
        $path = $1;
    } else {
        $path = 'TAINTED';
    }
    return $path;
}

# Subroutine: Usage
#
# Description: Prints a usage message.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#
#       1998/02/04: - Add '-m recipient' option.
#
#       1998/02/06: - Compression is configurable on a per file basis in
#                     the configuration file, so we no longer need that
#                     command line option.
#
sub Usage {
  my $progname = basename($0, ());

  print STDERR <<USAGE;
    Usage: $progname [-qdt] [-f <filename>] [-m <recipient>] frequency
USAGE

exit 1;
}

# Subroutine: MakeUniqueName
#
# Description: Creates a unique filename element using current date,
#              time, and process ID.
#
# Returns: A string containing the unique filename element
#
# Author: Sylvain Robitaille
#
# Date: 1998/04/08
#       1998/04/15: - Oops! Don't need to offset the day by one! :-)
#
#
sub MakeUniqueName {
  my @time_array  = localtime( time );
  my $unique_name = sprintf( "%04d%02d%02d.%02d%02d%02d.%d",
                             $time_array[5] + 1900,
                             $time_array[4] + 1,
                             $time_array[3],
                             $time_array[2],
                             $time_array[1],
                             $time_array[0],
                             $$ );
  return $unique_name;
}

# Subroutine: MakeErrorFile
#
# Description: Creates a file with a unique filename, ensuring that the
#              file doesn't already exist.
#
# Returns: A string containing the unique filename
#
# Author: Sylvain Robitaille
#
# Date: 2000/02/29
#
sub MakeErrorFile {
  my $errors_file;
  my $progname = shift;

  do {
    $errors_file = untaint_file "/tmp/$progname.errors." . MakeUniqueName;
    if ( $errors_file eq 'TAINTED' ) {
      fatal_error "Tainted errors filename: $errors_file";
    }
  } while ( -e $errors_file );

  # Now create the file, to be sure to shorten races as much as possible:
  open ERROR, "> $errors_file" or fatal_error "can't create $errors_file!";
  if (! chmod 0600, $errors_file ) {
    fatal_error "can't chmod $errors_file!";
  }
  close ERROR;

  # We can't use debug_message here, because it doesn't know yet where
  # $errors_file is:
  print "Error file created: $errors_file\n" if $debug;
  return $errors_file;
}

# Subroutine: Parse_Options
#
# Description: Parses the command line options into default global
#              variables.
#
# Uses globals: opt_d
#               opt_f
#               opt_m
#               opt_q
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#
#       1998/02/04: - Add '-m recipient' option to define who gets the
#                     mailed errors file.
#                   - We now check the return value of untaint file. If
#                     it's zero, the config file contains a tainted path
#                     and we cannot continue.
#
#       1998/02/06: - Compression is now configurable on a per file
#                     basis in the config file, so we don't need the
#                     command line option any more.
#                   - Changed check of return value from untaint_file to
#                     a string comparison, rather than checking for a
#                     numeric return value.
#
#       1998/02/17: - A more descriptive error message if config file
#                     path is tainted.
#
#       1998/02/19: - The mail recipient must also be untainted.
#
sub Parse_Options {

  if(! getopts('df:m:qt')) {
    Usage;
  }

  $debug       = $opt_d if( $opt_d );
  $quiet       = $opt_q if( $opt_q );
  $test_mode   = $opt_t if( $opt_t );

  $config_file = untaint_file $opt_f if( $opt_f );
  if( $config_file eq 'TAINTED' ) {
    fatal_error "Tainted path to config file: \'$opt_f\'";
  }

  $mail_rcpt   = untaint_mail $opt_m if( $opt_m );
  if( $mail_rcpt eq 'TAINTED' ) {
    fatal_error "Tainted email address: \'$opt_m\'";
  }
}

# Subroutine: Get_Pid( $pid_file )
#
# Description: Reads the contents of $pid_file, and ensures that it
#              contains a valid process id number.
#
# Notes: This used to be Read_Pid_file, but much error checking was
#        added. The routine now does its own complaining if we can't open
#        the file, providing a reasonably useful error message. This
#        routine also now checks that we can send a signal to the PID in
#        $pid_file, and complains if we can't.
#
# Returns: a valid process id, or zero if an error occurs.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/17: - Combine Read_Pid_File subroutine and sufficient
#                     error-checking.
#
#       1998/03/02: - Localize $. to avoid clobbering Read_Config's idea
#                     of it.
#
sub Get_Pid {
  my $pid_file = shift @_;
  my $pid;
  #
  # Let's make sure we don't clobber Read_Config's idea of $. (current
  # line number). We can localize it to ensure that.
  local $.;

  if(! open PID, $pid_file ) {
    error_message "Cannot open file $pid_file: $!";
    $pid = 0;
  } else {
    $pid = <PID>;
  }
  close PID;

  chomp $pid;
  #
  # The $pid should consist of only digits.
  if( $pid !~ /^(\d+)$/ ) {
    error_message "Malformatted pid in file $pid_file: $pid";
    $pid = 0;
  } else {
    #
    # Untaint the $pid while we're at it...
    $pid = $1;
  }
  #
  # Complain if we can't send this $pid a signal.
  if(! kill 'CONT', $pid ) {
    error_message "Unable to signal process $pid from file: $pid_file";
    $pid = 0;
  }
  return $pid;
}

# Subroutine: Read_Config( $config_file, $frequency )
#
# Description: Reads configuration data from $config_file into
#              appropriate arrays. Ignores items not matching
#              $frequency.
#
# Uses globals: $config_file
#               $errors_file
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02: - begin coding.
#
#       1998/02/04: - Changes to the config file format, order, and
#                     contents.
#                   - fields are separated with colons, optionally
#                     surrounded by whitespace.
#                   - we are now counting line numbers to provide more
#                     useful error messages.
#                   - missing $pid_path is no longer an error. It simply
#                     means that there is no process which needs to be
#                     HUPped in order to use the new log file.
#                   - We complain and skip an entry whose $backup_path
#                     is a symlink.
#
#       1998/02/05: - Untaint $backup_file.
#                   - Check that we can signal a $pid if one is given.
#
#       1998/02/06: - %backupfile array (hash) added.
#                   - %compress   array (hash) added.
#                   - $compress scalar now $compression.
#                   - We default to keeping an unlimited number of
#                     cycles if there is no $cycles field in the config
#                     file.
#                   - We determine whether or not we compress the backup
#                     file.
#                   - Fix offset by one bug when getting values out of
#                     the split line.
#                   - Fix bug checking for items that are defined but
#                     not true, (empty string).
#                   - if $compression is not empty string and not the
#                     word 'compress', that's an error. Complain and skip
#                     this file.
#                   - Increase the verbosity of some error messages to
#                     include line number and copy of line with all error
#                     messages.
#                   - Satisfy compiler warning with string comparison
#                     for if $pid != 'none'.
#                   - More checks: does $log_file exist? does $backup_dir
#                     exist? Once we know the extension, we want to be
#                     sure that $backup_dir/$backup_file.ext doesn't
#                     already exist, (either as a file or as a symlink).
#
#       1998/02/09: - Minor touches to organize the code a little
#                     better. Particularly the sections which check the
#                     $compress string and the $cycles setting.
#                   - We check to make sure that $backup_dir is rwx by us.
#
#       1998/02/10: - We don't need to check that we can read or write
#                     the $log_file, so long as we can write to the
#                     directory where it lives. That is checked in
#                     Move_Logs below.
#                   - We check that $cycles is a number, the string
#                     'date', or blank, though we don't act on it in any
#                     way yet.
#                   - bug fix: blank $cycles would have been rejected by
#                     not containing any digits.
#                   - actually, blank $cycles was not even correctly
#                     picked up from the config file. It (and any other
#                     adjacent blank fields) was not even processed in
#                     the split operation, causing a line with a blank
#                     $cycles field to always be skipped.
#
#       1998/02/12: - Using grep on the list of files in the backup
#                     directory to determine the highest existing $cycle,
#                     in the case of $cycle = '.' (keep infite number).
#                     This replaces code which searched for specific
#                     filename extensions, which of course is rather
#                     limiting.
#
#       1998/02/16: - Fix the call to grep. :-(
#
#       1998/02/17: - Fix a couple of error messages.
#                   - Don't need to count lines read from the config
#                     file. The current line number is available as $.
#       1998/02/18: - The frequency field in the config file can
#                     optionally be surrounded by whitespace. It must
#                     contain a colon.
#                   - We now complain if we read through the whole
#                     configuration file without processing any lines.
#                   - The fatal error raised if we can't open the
#                     configuration file has moved to the top of the if
#                     statement for easier reading: "if we can't open the
#                     config file to read, fatal error; else read through
#                     the file."
#                   - We no longer try and check the whole line for
#                     correct format. Rather, we check that mandatory
#                     entries are defined and complain/skip if they
#                     aren't. Optional entries are set to default values
#                     if they're not defined.
#                   - Bug fix: a tainted file path was overwritten with
#                     the string 'TAINTED', causing error messages to be
#                     less informative than they were intended to be.
#                   - Error messages now explicitly state that a line is
#                     being skipped.
#                   - The 'compress' field is case insensitive.
#                   - Paths should be absolute. This means it is an
#                     error if the first character of any path is not
#                     '/'.
#                   - The cycles field is mandatory. It *must* contain a
#                     number or the (case insensitive) string 'date'.
#                     There is no longer a way to save an infinite
#                     number of backups by number. This simply wasn't a
#                     feature for which there was a need and it added
#                     the possibility of confusion and/or clutter.
#                   - $backup_dir is tested first for a sym-link, then
#                     for a directory. We want to fail if it's a
#                     sym-link (even one which points to a directory),
#                     or if it isn't a directory.
#                   - We complain if an entry in the configuration file
#                     doesn't have the right number of entries, (assumes
#                     that the last entry will always be mandatory), or
#                     if the first field is missing. This will provide a
#                     rough syntax check on every eligible entry in the
#                     config file, (eligible means not a comment line and
#                     not a blank line), and (hopefully) identify errors
#                     before they cause file rotations to be missed.
#                   * This supersedes the previous change which had us
#                     checking only that each field in the line was
#                     defined and complaining if a mandatory field was
#                     missing, or setting optional fields to default
#                     values. If we have the right number of fields,
#                     optional fields are defined to the empty string.
#                   - The check for matching frequency is
#                     case-insensitive.
#
#       1998/02/19: - We fail on $log_file is a symlink or $backup_dir
#                     is a symlink iff we're defaulting $backup_dir to
#                     the path to $log_file, and either $log_file or its
#                     path is a symlink.
#
#       1998/02/25: - New 'date' keywords added: 'date-yesterday',
#                     'date-yestermonth' and 'date-time'
#
#       1998/03/02: - Get_Pid now localizes $. to avoid clobbering. This
#                     permits us to use $got_one for our own purposes,
#                     such as maybe counting how many lines we will
#                     process, etc.
#
#       1998/11/24: - Clean up several error messages.
#
#       2000/02/29: - local $. in GetPid doesn't seem to work on some
#                     systems (Perl-5.005?). Save the value before calling
#                     GetPid.
sub Read_Config {
  my $config_file = shift @_;
  my $frequency   = shift @_;
  my $got_one     = 0;
  my $cur_line;
  my @parts;
  my ($log_file, $backup_dir, $backup_file, $pid_file, $compression, $cycles);
  my $pid;
  my $untainted;
  my %have_pid;

  if(! open CONFIG, $config_file ) {
    fatal_error "Cannot open $config_file: $!";
  } else {
    debug_message "Reading $config_file";

    # read in the configuration
    while( <CONFIG> ) {
      next if /^\s*#/;                     # skip comment lines
      next unless /\w/;                    # skip blank lines

      chomp;                               # strip off newline
      #
      # Remove any leading or trailing whitespace from the line:
      $_ =~ s/^\s+//;
      $_ =~ s/\s+$//;

      @parts = split /\s*:\s*/, $_;
      #
      # If we don't have 7 parts ($#parts == 6), or if $parts[0] is the
      # empty string (''), we have a malformatted line.
      if( $#parts != 6 ) {
        error_message "Wrong number of fields in $config_file line $.: $_";
        error_message "Line $. is ignored";
        next;
      } elsif( $parts[0] eq '' ) {
        error_message "Missing frequency field in $config_file line $.: $_";
        error_message "Line $. is ignored";
        next;
      } elsif( lc $parts[0] ne lc $frequency ) {
        #
        # This line has at least the right number of parts, and the
        # first field is not empty. We can silently skip it if it
        # doesn't match our selected frequency.
        next;
      }
      #
      # Let's mark that we have a line worth processing.
      debug_message "got line: \"$_\"";
      $got_one++;
      #
      # let's get our settings from the parts list.
      (undef,
       $log_file,
       $backup_dir,
       $backup_file,
       $pid_file,
       $compression,
       $cycles) = @parts;
      #
      # Check to make sure we even have a log file name.
      if(! $log_file ) {
        error_message "Malformatted (missing log file field) entry".
                      " in $config_file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Make sure that the path to $log_file is absolute.
      if( $log_file !~ m!^/! ) {
        error_message "Relative path to log file: $log_file".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # We need to untaint the file paths which were provided. If any
      # path is tainted, we complain and skip the associated log file.
      $untainted = untaint_file $log_file;
      if( $untainted eq 'TAINTED' ) {
        error_message "Tainted path to log file: \'$log_file\'".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      } else {
        $log_file = $untainted;
      }
      #
      # We want to be sure the log file even exists, rather than try to
      # rotate non-existent files...
      #
      # The -f test returns true if the argument is a regular file, or a
      # sym-link which points to a regular file.
      if(! -f $log_file ) {
        error_message "File doesn't exist: $log_file".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # We also want to be sure that we can read this log file.
      if(! -r $log_file ) {
        error_message "Incorrect permission on log file: $log_file".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Place backup files in the same directory as $log_file if no
      # backup path.
      if(! $backup_dir ) {
        $backup_dir = dirname $log_file;
        #
        # We make sure that we are not going to default to a symlink,
        # either by $log_file being a symlink or its dirname being a
        # symlink.
        if( -l $log_file ) {
          error_message "Backup directory cannot default to $backup_dir".
                        " because $log_file is a symbolic link.";
          error_message "Skipping config file line $.: $_";
          next;
        }
        if( -l $backup_dir ) {
          error_message "Backup directory cannot default to $backup_dir".
                        " because $backup_dir is a symbolic link.";
          error_message "Skipping config file line $.: $_";
          next;
        }
      } else {
        #
        # Ensure that the path to the backup directory is absolute.
        if( $backup_dir !~ m!^/! ) {
          error_message "Relative path to backup directory: $backup_dir".
                        " from config file line $.: $_";
          error_message "Skipping line $.";
          next;
        }
        #
        # Ensure the path to the backup directory is untainted.
        $untainted = untaint_file $backup_dir;
        if( $untainted eq 'TAINTED' ) {
          error_message "Tainted path to backup directory: \'$backup_dir\'".
                        " from config file line $.: $_";
          error_message "Skipping line $.";
          next;
        } else {
          $backup_dir = $untainted;
        }
      }
      #
      # We want to be sure the backup directory even exists, rather than
      # try to put files in a non-existent directory.
      if(! -d $backup_dir ) {
        error_message "Directory doesn't exist: $backup_dir".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Check to make sure we have rw perms on $backup_dir.
      if(! -r $backup_dir || ! -w $backup_dir ) {
        error_message "Insufficient permission on backup directory:".
                      " $backup_dir".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Set $backup_file to basename of $log_file if $backup_file is
      # not defined, else untaint $backup_file.
      if(! $backup_file ) {
        $backup_file = basename $log_file;
      } else {
        $untainted = untaint_file $backup_file;
        if( $untainted eq 'TAINTED' ) {
          error_message "Tainted backup filename: \'$backup_file\'".
                        " from config file line $.: $_";
          error_message "Skipping line $.";
          next;
        } else {
          $backup_file = $untainted;
        }
      }
      #
      # If compression is defined, it should be the string 'compress'.
      # Anything else is an error.
      if( $compression && lc $compression ne 'compress' ) {
        error_message "Incorrect compression field: $compression".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Get the PID of process owning this file if appropriate.
      if(! $pid_file) {
        $pid = 'none';
      } else {
        #
        # Check to make sure that path to $pid_file is absolute.
        if( $pid_file !~ m!^/! ) {
          error_message "Relative path to PID file: $pid_file".
                        " from config file line $.: $_";
          error_message "Skipping line $.";
          next;
        }
        #
        # Untaint the path to $pid_file.
        $untainted = untaint_file $pid_file;
        if( $untainted eq 'TAINTED' ) {
          error_message "Tainted path to PID file: \'$pid_file\'".
                        " from config file line $.: $_";
          error_message "Skipping line $.";
          next;
        } else {
          $pid_file = $untainted;
        }
        #
        # If we already read this $pid_file, we don't need to read it
        # again.
        if( defined $have_pid{$pid_file} ) {
          $pid = $have_pid{$pid_file};
        } else {
          #
          # Read the PID from $pid_file.
          $cur_line = $.;
          $pid = Get_Pid $pid_file;
          #
          # skip this entry if we can't read the $pid_file
          if( $pid == 0 ) {
            error_message "There was an error getting the PID".
                          " from file $pid_file".
                          " from config file line $cur_line: $_";
            error_message "Skipping line $cur_line";
            next;
          }
        }
        #
        # Make a note that we already have the pid from this file.
        $have_pid{$pid_file} = $pid;
      }
      #
      # We check that $cycles is defined. The entry is malformatted if
      # doesn't contain a $cycles field.
      if(! $cycles ) {
        error_message "Malformatted (missing cycles field) entry".
                      " in $config_file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # Syntax check on $cycles. This should be a number, or one of the
      # following (case insensitive) strings: 'date', 'date-yesterday',
      # 'date-yestermonth', 'date-time'.
      if(    $cycles !~ /^\d+$/            &&
          lc $cycles ne 'date'             &&
          lc $cycles ne 'date-yesterday'   &&
          lc $cycles ne 'date-yestermonth' &&
          lc $cycles ne 'date-time' ) {
        error_message "Incorrect cycles field: $cycles".
                      " from config file line $.: $_";
        error_message "Skipping line $.";
        next;
      }
      #
      # ok so now build this into the arrays.
      push @{$logfiles{$pid}}, $log_file;
      $rotation{$log_file}   = $cycles;
      $backupdir{$log_file}  = $backup_dir;
      $backupfile{$log_file} = $backup_file;
      $compress{$log_file}   = $compression;
    }
    close CONFIG;
    #
    # Make sure we found at least one matching entry in the config file.
    if(! $got_one ) {
      error_message "No entries found matching $frequency in $config_file";
    }
  }   # if ! open $config_file
}   # Read_Config

# Subroutine: Move_Logs
#
# Description: This routine goes through each log file in the %logfiles
#              array, and moves the file aside to create a new log file.
#              It then creates the new log file, and HUP's the
#              associated process if applicable.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/09: - begin coding, per Anne's spec.
#
#       1998/02/10: - Use rename to move new temp file into place
#                     without risk of ever losing the existing file if
#                     disaster strikes. Much checking along the way.
#                   - Create a list of files which have successfully
#                     been moved aside and replaced. The old files will
#                     be dealt with later.
#
#       1998/02/11: - Clarify error message if we can't HUP a process.
#                     More than one file might be left in place.
#
#       1998/02/16: - Clarify error message if we can't create a hard
#                     link between $realfile and $realfile.rotate-old.
#
#       1998/02/18: - Rewrite the loops which unravel sym-links into one
#                     better organized loop. We're already fairly certain
#                     that we have a real file at the other end, but
#                     let's still make sure that we can't go infinitely.
#                     If we go through 100 times, we call that too many
#                     and fail.
#                   - The organization of the new loop was suggested by
#                     Anne. Any bugs within are my own. :-)
#                   - We no longer change to the directory. Once we
#                     unravel the sym-links, we know that we have a real
#                     file in a real directory.
#                   - Tests after chown/chmod are separated. We do both
#                     chown (with uid and gid) and chmod, and if either
#                     fails, we try and determine precisely why.
#                   - Some clarification in comments.
#                   - The %done_files and @skip_rotate arrays are no
#                     longer required. If we must skip rotating a file,
#                     it simply is removed from the
#                     @{$logfiles{$this_pid}} list. If that list empties
#                     or we cannot HUP $this_pid, the hash entry for
#                     $logfiles{$this_pid} is deleted. This leaves us
#                     with a %logfiles hash containing only files which
#                     we are to proceed with.
#
#       1998/02/19: - Bug fix: The chmod of the new file was not working
#                     properly.
#
#       1998/02/20: - Bug fix: Logic error when testing if chown and chmod
#                     of the new file were successful.
#
#       1998/02/22: - Bug fix: error in call to kill HUP. :-(
#                   - Bug fix: incorrect grouping when checking chmod/chown.
#
#       1998/03/02: - Bug fix: error following relative directory
#                     sym-links.
#                   - $f_inode should not be constantly re-declared in
#                     the foreach block. Declare once with other file
#                     stat variables.
#
#       1998/11/24: - Clean up several error messages.
#       1999/02/01: - Bug fix: relative symlinks mishandled, and
#                     symlinked log files not correctly rotated.
#       2001/04/20: - Bug fix: add taint checks to readlink results.
#                     Simplified the "break out when too many
#                     symlinks" code using a label on the foreach
#                     statement.
#
sub Move_Logs {
  #
  # From global arrays.
  my $this_pid;
  my $this_file;
  #
  # Used during sym-link unravelling.
  my $givendir;
  my $givenfile;
  my $realdir;
  my $realfile;
  my $loop_count;
  #
  # File stats.
  my ($dev, $ino, $mode, $uid, $gid, $f_ino, $f_mode, $f_uid, $f_gid);

  foreach $this_pid (keys %logfiles) {
    PROCESS_FILES:
    foreach $this_file (@{$logfiles{$this_pid}}) {
      #
      # reset our loop counter.
      $loop_count = 0;
      #
      # name        path     suffix
      ($givenfile, $givendir, undef) = fileparse( $this_file, '');
      #
      # We will want to check the file by its fullpath.
      $realdir  = $givendir;
      $realfile = $this_file;
      #
      # We need to strip off the trailing / if we want to be able to
      # determine that we have a sym-link.
      $realdir =~ s%/$%%;
      #
      # This is the loop to unravel any sym-links and point us to a real
      # file in a real directory.
      while( -l $realfile || -l $realdir ) {
        #
        # Make sure we're not going infinitely.
        $loop_count++;
        if( $loop_count > 100 ) {
          error_message "Too many links while trying to process $this_file.";
          error_message "Skipping file $this_file.";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
          next PROCESS_FILES;
        }
        #
        # If $realdir is a sym-link, we follow the links until we have
        # a real directory.
        #
        while( -l $realdir ) {
          $realdir = readlink $realdir;
          {
            my $untainted = untaint_file $realdir;
            if( $untainted eq 'TAINTED' ) {
              error_message "Tainted path in symlink contents: $realdir".
                            " Parsed from log file name $this_file";
              error_message "Skipping file $this_file.";
              @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                              @{$logfiles{$this_pid}};
              next PROCESS_FILES;
            } else {
              $realdir = $untainted;
            }
          }
          #
          # Now if the link is relative, add the path portions of
          # $givendir back.
          if( $realdir !~ m%^/% ) {
            #
            # strip off the last element of $givendir.
            $givendir = dirname $givendir;
            $realdir  = "$givendir/$realdir";
          }
          #
          # Now reset $givendir to $realdir.
          $givendir = $realdir;
          #
          # And again make sure that we check correctly for a sym-link.
          $realdir  =~ s%/$%%;
          #
          # We need to be sure that $givendir has a trailing / for when
          # we need to prepend it.
          if( $givendir !~ m%/$% ) {
            $givendir = $givendir.'/';
          }
          $realfile = $givendir.$givenfile
        }
        #
        # ok, so now we're looking at a real directory which contains
        # $givenfile. We want to be sure that the file is not a
        # sym-link, or else we start over.
        if( -l $realfile ) {
          $realfile = readlink $realfile;
          {
            my $untainted = untaint_file $realfile;
            if( $untainted eq 'TAINTED' ) {
              error_message "Tainted path in symlink contents: $realfile".
                            " Parsed from log file name $this_file";
              error_message "Skipping file $this_file.";
              @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                              @{$logfiles{$this_pid}};
              next PROCESS_FILES;
            } else {
              $realfile = $untainted;
            }
          }
          ($givenfile, $givendir, undef) = fileparse $realfile;
          $givendir =~ s%/$%%;

          if( $givendir =~ m%^/% ) {
            #
            # The sym-link is absolute
            $realdir = $givendir;
          } else {
            #
            # relative sym-link
            $realdir  = "$realdir/$givendir";
            $realfile = "$realdir/$givenfile";
          }
        }
      }
      #
      # We already know that we have correct permissions on $givenfile,
      # but we need rw perms on $givendir, so check that.
      if( ! -r $realdir || ! -w $realdir ) {
        #
        # NOTE that we no longer know which line of the config file this
        #      comes from, so we are unable to provide as clear an error
        #      message.
        error_message "Incorrect permissions on directory: $realdir".
                      " Parsed from log file name $this_file";
        error_message "Skipping file $this_file.";
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
        next;
      }
      #
      # From here on, we're working with $realfile in $realdir, which may
      # be different from $this_file:
      debug_message "real dir = $realdir; real file = $realfile";

      # Proceed from here only if we're not in test_mode:
      if($test_mode) {
        info_message "TEST_MODE: not rotating $realfile";
      } else {
      #
      # We need some stat info from the $realfile: ownership, perms,
      # inode and filesystem.
      ($dev, $ino, $mode, undef, $uid, $gid, undef ) = stat $realfile;
  
      #
      # Now we make a hard link to the file. Note that this will fail if
      # the link we're trying to create already exists.
      if(! link $realfile, "$realfile.rotate-old" ) {
        error_message "Cannot create link $realfile.rotate-old: $!";
        error_message "Skipping file $this_file.";
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
        next;
      }
      #
      # And we create a new file to use:
      #
      # First make sure there isn't one already:
      if( -e "$realfile.rotate-new" ) {
        error_message "File $realfile.rotate-new already exists!";
        error_message "Skipping file $this_file.";
        unlink "$realfile.rotate-old";
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
        next;
      }
      #
      # All clear:
      if(! open NEWFILE, ">$realfile.rotate-new" ) {
        error_message "Cannot create file $realfile.rotate-new: $!";
        error_message "Skipping file $this_file.";
        unlink "$realfile.rotate-old";
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
        next;
      }
      close NEWFILE;
      #
      # Now we want to be sure to properly set up ownership and perms on
      # the new file:
      if(! ((chmod $mode, "$realfile.rotate-new") &&
            (chown $uid, $gid, "$realfile.rotate-new")) ) {
  
        (undef, undef, $f_mode, undef, $f_uid, $f_gid, undef) =
            stat "$realfile.rotate-new";
        #
        # Chown will fail if we're not root. This is fine as long as the
        # file ownership is the same as the file we're replacing.
        if( $f_uid != $uid ) {
          error_message "Cannot chown file $realfile.rotate-new!";
          error_message "Skipping file $this_file.";
          unlink "$realfile.rotate-old";
          unlink "$realfile.rotate-new";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
          next;
        } elsif( $f_gid != $gid ) {
          error_message "Cannot chgrp file $realfile.rotate-new!";
          error_message "Skipping file $this_file.";
          unlink "$realfile.rotate-old";
          unlink "$realfile.rotate-new";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
          next;
        } elsif( $f_mode != $mode ) {
          error_message "Cannot chmod file $realfile.rotate-new!";
          error_message "Skipping file $this_file.";
          unlink "$realfile.rotate-old";
          unlink "$realfile.rotate-new";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
          next;
        }
      }
      #
      # Now we rename $realfile.rotate-new to $realfile. This will
      # ensure that no matter what happens, $realfile will always
      # exist. At this point, it is expected that logging continues to
      # be sent to $realfile.rotate-old.
      #
      # We complain and skip this file if for some reason we cannot
      # complete the operation. We clean up $realfile.rotate-old iff
      # $realfile still exists with the same inode number as before.
      #
      if( ! rename "$realfile.rotate-new", $realfile ) {
        error_message "Cannot rename $realfile.rotate-new to".
                      " $realfile: $!";
        error_message "Skipping file $this_file.";
        unlink "$realfile.rotate-new";
        (undef, $f_ino, undef) = stat $realfile;
        if( $f_ino == $ino ) {
          error_message "Cleaning up $realfile.rotate-old";
          unlink "$realfile.rotate-old";
        } else {
          error_message "$realfile has new inode; leaving ".
                        "$realfile.rotate-old in place.";
          error_message "MANUAL CLEANUP REQUIRED!";
        }
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                         @{$logfiles{$this_pid}};
        next;
      }
      # Now add the real file to the list of real files.
      #
      $realfiles{$this_file} = $realfile;
    }  # if ! $test_mode
    }   # foreach $this_file
    #
    # Now that we've moved aside all files related to $this_pid, we want
    # to HUP the process. Complain if we can't, and ensure that we won't
    # do anything more with @{$logfiles{$this_pid}}.
    #
    # We don't need to do this if @{$logfiles{$this_pid}} no longer
    # contains any elements, in which case, we can simply delete the
    # hash entry for $logfiles{$this_pid}.
    if(! @{$logfiles{$this_pid}} ) {
      delete $logfiles{$this_pid};
    } elsif(($this_pid ne 'none') && ! (kill 'HUP', $this_pid)) {
      error_message "Cannot HUP process $this_pid: $!".
                    " Files will be left in place.";
      error_message "MANUAL CLEANUP REQUIRED!";
      delete $logfiles{$this_pid};
    }
  }   # foreach $this_pid
}   # Move_Logs

# Subroutine: Rotate_Logs
#
# Description: This is the routine which moves old log files around, if
#              necessary, and place the current backup in the
#              appropriate directory.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/12: - Make sure we don't process files which are
#                     associated with pid's which we could not HUP.
#                   - Change to the backup directory. Complain and skip
#                     if we can't.
#
#       1998/02/16: - Rotate the old logs if necessary.
#                   - Create a link to (or a copy of) the temp copy of
#                     the "current" backup file. 
#
#       1998/02/18: - We no longer check through a list of PIDs to see
#                     if we should skip rotation. Instead, we can be
#                     certain that our @{$logfiles{$this_pid}} list
#                     contains only files which were sucessfully moved
#                     aside. The keys to $logfiles{$this_pid} are only
#                     PIDs which were successfully HUPped.
#                   - We no longer chdir to the backup directory. We
#                     already complained when we were reading the config
#                     file if $backup_dir was a sym-link.
#                   - rewrite of the bit that gets the "old cycle" from
#                     a file being rotated, increments it and forms a
#                     new name for that file. It is now done by reg-exps
#                     rather than trying to parse the filename.
#                   - We now unlink $this_file.rotate-old right away if
#                     we have successfully made a copy in the
#                     $backup_dir.
#                   - There is no longer a list of %rotated files. We
#                     simply remove files from @{$logfiles{$this_file}}
#                     if we have trouble rotating.
#
#       1998/02/19: - Complete error/progress checking. Make sure we
#                     don't continue rotating a file if an error occurs
#                     along the way, (ie.: can't properly rotate a
#                     previous backup).
#                   - Some of the code is re-organized.
#                   - Bug fix: We complain/skip if the file we're trying
#                     to backup to already exists.
#                   - Bug fix: $cycles should be case insensitive.
#                   - Bug fix: regular expressions to find all backup
#                     files were wrong.
#
#       1998/02/20: - Date adjustment. We don't need yesterday's date
#                     because we're actually running before midnight.
#
#       1998/02/25: - New 'date' keywords added: 'date-yesterday',
#                     'date-yestermonth' and 'date-time'.
#
#       1998/03/02: - Bug fix: Incorrect error message if cp fails.
#                     (missing $).
#
#       1998/09/08: - Additional taint-checking. As we read in filenames
#                     from the backup_dir directory.
#
#       1998/09/15: - Bug fix: Include the '$!' error string in error
#                     message if rename fails
#
#       1998/09/18: - Bug fix: Initialize our list of old_backups before
#                     adding entries from the directory listing; (Now
#                     that we're untainting those, we're only adding
#                     them one at a time, rather than as an entire list)
#                   - Also initialize our taint-checking variable. The
#                     directory we're backing up to doesn't necessarily
#                     contain any old files, resulting in possible errors
#                     (or use of an unitialized value).
#
#       1998/11/24: - Clean up several error messages.
#                   - We capture the output (if any) of the cp command,
#                     and print it along with our own error message.
#
sub Rotate_Logs {
  #
  # data from the global arrays.
  my $this_pid;
  my $this_file;
  my $backup_dir;
  my $backup_file;
  my $cycle;
  #
  # Local workers.
  my @old_backups;
  my @files;
  my $file;
  #
  # Filename parsing.
  my $name;
  my $ext;
  my $suffix;
  #
  # Date values.
  my $year;
  my $mon;
  my $day;
  my $hour;
  my $min;
  my $sec;
  #
  # for untainting
  my $untainted;
  #
  # For dealing with output of system commands
  my $output;

  foreach $this_pid (keys %logfiles) {
    foreach $this_file (@{$logfiles{$this_pid}}) {
      $backup_file = $backupfile{$this_file};
      $cycle       = $rotation{$this_file};
      $backup_dir  = $backupdir{$this_file};

      if( lc $cycle !~ m/^date/ ) {
        #
        # we need a list of files to rotate, since there may be several
        # different types of file extensions to consider.
        if( ! opendir BACKUPDIR, $backup_dir ) {
          error_message "Can't read from dir $backup_dir: $!".
                        " Skipping rotation for $this_file.";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
          next;
        }
        #
        # We already untainted $backup_file when we read it from the
        # config file, but it seems as though newer versions of Perl
        # will check this more strictly. Since we are reading in a
        # directory listing here, we need to untaint each file we read
        # in, or else Perl complains when we try and rotate older
        # backups.
        #
        # first initialize our variables:
        @old_backups = ();
        $untainted = "";

        foreach $file (grep /^($backup_file)/, readdir BACKUPDIR) {
          $untainted = untaint_file $file;
          if( $untainted eq 'TAINTED' ) {
            error_message "Tainted filename in $backup_dir: \'$file\'".
            error_message "Skipping file $this_file.";
            @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                             @{$logfiles{$this_pid}};
            last;
          } else {
            debug_message "Adding $untainted to list of files to rotate.";
            push @old_backups, $untainted;
          }
        }
        closedir BACKUPDIR;
        #
        # Now actually skip this file if we found a tainted filename
        next if $untainted eq 'TAINTED';

        # Now we rotate from the oldest to the newest.
        #
        # Start by removing the oldest "max" $cycles files.
        $name = "$backup_file.$cycle";
        @files = grep /^($name)(\..*)*$/, @old_backups;
        foreach $file (@files) {
          debug_message "Unlinking $backup_dir/$file";
          if(! unlink "$backup_dir/$file" ) {
            error_message "Unable to unlink file: $backup_dir/$file".
                          ". Aborting rotation for $this_file.";
            error_message "MANUAL CLEANUP REQUIRED!";
            @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
            last;
          }
          #
          # Remove this $file from the list of @old_backups to rotate.
          @old_backups = grep !/^($file)$/, @old_backups;
        }
        #
        # We skip to the next file if the %logfiles hash no longer
        # contains $this_file.
        if(! grep /^($this_file)$/, @{$logfiles{$this_pid}} ) {
          next;
        }
        #
        # Now we work our way back to the most recent backup.
        while( $cycle ) {
          $name = "$backup_file.$cycle";
          @files = grep /^($name)(\..*)*$/, @old_backups;
          #
          foreach $file (@files) {
            # filename, cycles, optional other.
            $file   =~ m!(.+)\.(\d+)(.*)!;
            $name   =  $1;
            $ext    =  $2;
            $suffix =  $3;
            $ext++;             # increment for new filename
            #
            # Now rotate.
            debug_message "Renaming $backup_dir/$file to ".
                                   "$backup_dir/$name.$ext$suffix";
            if(! rename "$backup_dir/$file",
                        "$backup_dir/$name.$ext$suffix" ) {
              error_message "Could not rename $backup_dir/$file to".
                            " $backup_dir/$name.$ext$suffix: $!";
              error_message "Aborting rotation for $this_file.";
              error_message "MANUAL CLEANUP REQUIRED!";
              @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                             @{$logfiles{$this_pid}};
              last;
            }
          }   # foreach $file
          #
          # We break out of the while $cycles loop if the %logfiles hash
          # no longer contains $this_file.
          if(! grep /^($this_file)$/, @{$logfiles{$this_pid}} ) {
            last;
          }
        } continue {
          $cycle--;
        } # while $cycle
        #
        # We skip to the next file if the %logfiles hash no longer
        # contains $this_file.
        if(! grep /^($this_file)$/, @{$logfiles{$this_pid}} ) {
          next;
        }
        $ext = '1';
      } else {  # $cycle =~ m/^date/
        #
        # If it's date-yesterday, we just use yesterday's date.
        if( lc $cycle eq 'date-yesterday' ) {
          (undef, undef, undef, $day, $mon, $year, undef) =
            localtime( $time_at_start - (60 * 60 * 24) );
          #
          # Adjust $year to full value, and adjust $mon from 0 = January.
          $year += 1900;
          $mon  += 1;
          $ext = sprintf "%04d%02d%02d", $year, $mon, $day;
        } else { # $cycle ne 'date-yesterday'
          #
          # We use today's date and adjust as necessary.
          ($sec, $min, $hour, $day, $mon, $year, undef) =
            localtime( $time_at_start );
          # adjust $year to full value:
          $year += 1900;
          #
          # if we're using 'date-yestermonth', we don't increment $month
          # by one, but we do adjust to 12 and decrement year if it's 0.
          if( lc $cycle eq 'date-yestermonth' ) {
            if( $mon == 0 ) {
              $year--;
              $mon = 12;
            }
            $ext = sprintf "%04d%02d", $year, $mon;
          } else { # $cycle ne 'date-yestermonth'
            #
            # Other date keywords want _this_ month, so adjust it from
            # 0 = January to 1 = January.
            $mon  += 1;
            $ext = sprintf "%04d%02d%02d", $year, $mon, $day;
          }
          #
          # If we're using 'date-time', we append ".HHMMSS" to the
          # extension.
          if( lc $cycle eq 'date-time' ) {
            $ext .= sprintf ".%02d%02d%02d", $hour, $min, $sec;
          }
        }   # if $cycle eq 'date-yesterday'
      }   # if cycle !~ m/^date/
      #
      # Now we can move $realfiles{$this_file}.rotate-old to
      # $backup_dir/$backup_file.$ext
      #
      # We want to fail if $backup_dir/$backup_file.$ext already exists.
      if( -e "$backup_dir/$backup_file.$ext" ) {
        error_message "cannot backup $this_file:".
                      " $backup_dir/$backup_file.$ext already exists!";
        error_message "MANUAL CLEANUP REQUIRED!";
        @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
        next;
      }   # if file already exists.
      #
      # We can try link first and then if that fails (files are not
      # on the same filesystem, we can try copying and removing the
      # old file.
      if(! link "$realfiles{$this_file}.rotate-old",
                "$backup_dir/$backup_file.$ext" ) {
        #
        # If we couldn't link, we're probably dealing with separate file
        # systems, so we need to copy the file over. Let's make sure to
        # capture the output.
        $output = `cp -p $realfiles{$this_file}.rotate-old $backup_dir/$backup_file.$ext 2>&1`;
        chomp $output;

        if( $output ) {
          #
          # If there was an error, extract the message, and print it along
          # with our own error message.
          $output =~ s%.*: %%;
          error_message "Error copying file from ".
                        "$realfiles{$this_file}.rotate-old to ".
                        "$backup_dir/$backup_file.$ext: $output";
          error_message "Leaving $this_file.rotate-old in place!";
          error_message "MANUAL CLEANUP REQUIRED!";
          @{$logfiles{$this_pid}} = grep !/^($this_file)$/,
                                           @{$logfiles{$this_pid}};
        }   # if an error occurred while copying the file.
      }   # if we couldn't create a hardlink.
      #
      # We skip to the next file if the %logfiles hash no longer
      # contains $this_file.
      if(! grep /^($this_file)$/, @{$logfiles{$this_pid}} ) {
        next;
      }
      #
      # We can safely unlink $this_file.rotate-old bcause we know we
      # have a good copy of it.
      unlink "$realfiles{$this_file}.rotate-old";
      if( -e "$realfiles{$this_file}.rotate-old" ) {
        error_message "Error removing file $this_file.rotate-old!";
        error_message "MANUAL CLEANUP REQUIRED!";
      }
      #
      # Let's update the %backupfile hash to contain the full real
      # path to the backup we just created. This will be used by the
      # Do_Compress subroutine to know which file to compress, if
      # appropriate.
      $backupfile{$this_file} = "$backup_dir/$backup_file.$ext";
      #
      # Announce the progress.
      info_message "$frequency rotation complete: $this_file";
    }   # foreach $this_file
  }   # foreach $this_pid
}   # Rotate_Logs


# Subroutine: Do_Compress
#
# Description: Compresses the backup files if appropriate.
#
# Note: This routine will only work with log files which were
#       successfully rotated in the Rotate_Logs subroutine.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/16: - begin code
#
#       1998/02/18: - There is no more %rotated hash. The %logfiles hash
#                     now contains only files which have been
#                     successfully rotated, and we can work with them.
#
#       1998/02/19: - We now get our filename to backup from the
#                     %backupfile hash.
#                   - We don't need to try and decipher what went wrong
#                     if the call to gzip fails. What we do need is to
#                     see if a failed gzip file is left behind and
#                     remove it.
#                   - Bug fix: We need to iterate through each file in
#                     each list associated with each pid.
#
#       1998/11/24: - We capture the output (if any) of the gzip command,
#                     and print it along with our own error message.
#
#       2001/04/20: - Bug fix: detect an existing gz file *before*
#                     trying to compress a log file.
#
sub Do_Compress {
  my $this_pid;
  my $this_file;
  my $backup_file;
  my $output;

  foreach $this_pid (keys %logfiles) {
    foreach $this_file (@{$logfiles{$this_pid}}) {
      $backup_file = $backupfile{$this_file};
  
      if( $compress{$this_file} ) {
        debug_message "Compressing $backup_file with gzip.";
        #
        # Check that we're not about to overwrite an existing
        # compressed file.
        if( -e "$backup_file.gz" ) {
          error_message "Error attempting to compress $backup_file: $backup_file.gz exists";
          error_message "MANUAL CLEANUP REQUIRED!";
        } else {
          #
          # Do the compression, but capture any output from the command.
          $output = `gzip --best $backup_file 2>&1`;
          chomp $output;
  
          #
          # If there was an error, extract the message and print it along
          # with our own error message, otherwise report success.
          if( $output ) {
            $output =~ s%.*: %%;
            error_message "Error attempting to compress $backup_file: $output";
            if( -f "$backup_file.gz" ) {
              if(! unlink "$backup_file.gz" ) {
                error_message "Unable to unlink stray $backup_file.gz";
                error_message "MANUAL CLEANUP REQUIRED!";
              }   # If we couldn't unlink
            }   # If there was a stray
          } else {
            info_message "Compress $backup_file: done.";
          }   # If an error occured while gzipping (else)
        }   # If a .gz file existed before we started (else)
      }   # if we're supposed to compress the backup
    }   # foreach $this_file
  }   # foreach $this_pid
}   # Do_Compress


# The main routine.
#
# Author: Sylvain Robitaille
#
# Date: 1998/02/02
#
#       1998/02/04: - Update to include logging to syslog.
#
#       1998/02/17: - info_message no longer needs the program name,
#                     since that is already logged by syslog.
#
#       1998/03/02: - Remove "test initiated" syslog message. Now log
#                     "log rotations initiated."
#

# Variables used by main.
my $facility;        # Which facility will we log to syslog with?
my $progname;        # Who are we, anyway?

# Now get command line options.
Parse_Options;

# The remaining option should be the frequency.
if($#ARGV != 0) {
  Usage;
}
$frequency = shift @ARGV;

# Who are we, anyway?
$progname = basename $0;

# Let's get the time at which we started, to be used for creating
# date-based filename extensions.
$time_at_start = time;

# Ok we're started up correctly. We want to open up a connection to
# syslog, with the appropriate facility, and log that we're starting.
#
# We have the following options for syslog logging:
#
# pid:    Logs the process ID with each message.
# ndelay: Opens the connect to syslogd immediately, instead of when the
#         first message is logged.
# cons:   Sends messages to the console if unable to send them to
#         syslogd.
# nowait: Logs messages to the console without waiting for child
#         processes that are forked.
#
# $> is our effective user id.
if( $> == 0 ) {
  $facility = $root_syslog_facility;
} else {
  $facility = $user_syslog_facility;
}
openlog $progname, 'pid', $facility;

# Ok, here we go...
info_message "$frequency log rotations initiated.";

# Absolute path to the errors file. So far, an errors file is created
# whether or not it will be mailed to anyone. The file is subsequently
# deleted at the end of the cycle.
#
# 2000/02/29: This file is now created automatically, in order to reduce
#             the risk of a sym-link attack based on predictable filenames.
$errors_file = MakeErrorFile $progname;

# Read the config file.
Read_Config( $config_file, $frequency );

# Now move the existing logs aside, and replace them, HUPping associated
# processes where required.
Move_Logs;

if(!$test_mode) {
# Now we can rotate the old log files around, and move the "current old
# one" to the backup directory.
Rotate_Logs;

# Compress those files which were configured to do so.
Do_Compress;
}

if( -s $errors_file ) {
  info_message "$frequency log rotations done; ERRORS.";
  #
  # Mail the errors file to $mail_rcpt if $mail_rcpt is defined.
  # Output to STDERR if we can't.
  open ERRORS, $errors_file;
  if( $mail_rcpt ) {
    #
    # We need to untaint the $mail_rcpt before we're permitted to use
    # it.
    if(! open MAILER, "| $mail_program") {
      print STDERR "Unable to start mail program '$mail_program': $!\n",
                   <ERRORS>;
    } else {
      print MAILER "To: $mail_rcpt\n",
                   "Subject: $subject\n",
                   "\n";
      print MAILER <ERRORS>;
      close MAILER or
        print STDERR "Error closing mail program '$mail_program': $!\n",
              <ERRORS>;
    }
  }
  close ERRORS;
} else {
  info_message "$frequency log rotations done; ok."
}

unlink $errors_file;
closelog;

